import importlib
import platform
import sys
from argparse import ArgumentParser
from typing import Iterable, List, Optional, Sequence, Set, Tuple

from . import __version__
from .config import Config
from .linter import Linter
from .problem import ProblemType

IGNORABLE_TYPES = [t.name for t in (ProblemType.URI, ProblemType.FLOAT)]


def main(argv: Optional[Sequence[str]] = None) -> int:
    if argv is None:
        argv = sys.argv[1:]

    parser = _make_arg_parser()
    args = parser.parse_args(argv)
    config = Config(args.ignore, args.yaml)

    if config.is_yaml and not _is_pyyaml_installed():
        print("Missing optional dependency 'PyYAML'.", file=sys.stderr)
        return 2

    all_problems = dict(_get_problems(args.filepath_or_text, config))

    if not all_problems:
        return 0

    printing_input = len(args.filepath_or_text) > 1
    for key, problems in all_problems.items():
        if printing_input:
            print(key)

        problems_count = len(problems)
        header = "One error" if problems_count == 1 else f"{problems_count} errors"
        print(f"{header}:")

        for problem in problems:
            print(f" {problem}")

    return 1


def _get_problems(
    input_list: List[str], config: Config
) -> Iterable[Tuple[str, List[str]]]:
    for filepath_or_text in input_list:
        problems = Linter.validate(filepath_or_text, config)
        if problems:
            yield filepath_or_text, problems


def _is_pyyaml_installed() -> bool:
    try:
        importlib.import_module("yaml")
    except ImportError:
        return False
    return True


def _make_arg_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument(
        "filepath_or_text",
        nargs="+",
        help="ASL file path or ASL JSON/YAML text.",
    )

    parser.add_argument(
        "--ignore",
        type=_ignore_types,
        help="Comma-separated list of errors to ignore (or skip)."
        " For example, `--ignore=URI,FLOAT`."
        f" Supported values are {IGNORABLE_TYPES}.",
    )

    parser.add_argument(
        "--yaml",
        action="store_true",
        help="Parse as YAML instead of JSON.",
    )

    parser.add_argument(
        "--version",
        "-V",
        action="version",
        version=f"{__version__} {_get_python_version()}",
    )
    return parser


def _ignore_types(value: str) -> Set[ProblemType]:
    values = set(v.strip() for v in value.split(","))
    if values - set(IGNORABLE_TYPES):
        raise ValueError()
    return set(ProblemType[v] for v in values)


def _get_python_version() -> str:
    return (
        f"{platform.python_implementation()} "
        f"{platform.python_version()} on {platform.system()}"
    )



import operator
from abc import ABC, abstractmethod
from typing import Any, List, Optional, Tuple, Union

from ..problem import ProblemPredicate, ProblemType
from .base import Field, NonNullMixin


class BaseNumericField(NonNullMixin, Field, ABC):
    _floor: Optional[int]
    _ceiling: Optional[int]

    def __init__(
        self,
        name: str,
        floor: Optional[int] = None,
        ceiling: Optional[int] = None,
        inclusive: bool = True,
    ) -> None:
        super().__init__(name)
        self._floor = floor
        self._ceiling = ceiling
        self._inclusive = inclusive

    @property
    @abstractmethod
    def problem_type(self) -> ProblemType:
        pass  # pragma: no cover

    @property
    @abstractmethod
    def raw_type(self) -> Union[type, Tuple[type, ...]]:
        pass  # pragma: no cover

    def validate(self, value: Any) -> List[ProblemPredicate]:
        problems = super().validate(value)
        if problems:
            return problems
        problems = self.check_type(value, self.raw_type, self.problem_type)
        if problems:
            return problems

        ope = operator.le if self._inclusive else operator.lt
        if isinstance(self._floor, int) and ope(value, self._floor):
            return [ProblemPredicate(f" is {value} but allowed floor is {self._floor}")]
        ope = operator.ge if self._inclusive else operator.gt
        if isinstance(self._ceiling, int) and ope(value, self._ceiling):
            return [
                ProblemPredicate(f" is {value} but allowed ceiling is {self._ceiling}")
            ]
        return []


class IntegerField(BaseNumericField):
    @property
    def problem_type(self) -> ProblemType:
        return ProblemType.INTEGER

    @property
    def raw_type(self) -> Union[type, Tuple[type, ...]]:
        return int


class FloatField(BaseNumericField):
    @property
    def problem_type(self) -> ProblemType:
        return ProblemType.FLOAT

    @property
    def raw_type(self) -> Union[type, Tuple[type, ...]]:
        return float


class NumericField(BaseNumericField):
    @property
    def problem_type(self) -> ProblemType:
        return ProblemType.NUMERIC

    @property
    def raw_type(self) -> Union[type, Tuple[type, ...]]:
        return int, float



import json
import os.path
from typing import Any, Callable, Dict, List, Optional, Union

from .config import Config
from .nodes.factory.factory import StateFactory
from .nodes.state_machine import StateMachine


class ParseError(ValueError):
    def __init__(self, root_cause: Exception) -> None:
        super().__init__()
        self.root_cause = root_cause


class Linter:
    @staticmethod
    def validate(
        json_str_or_dict_or_file_path: Union[str, Dict[str, Any]],
        config: Optional[Config] = None,
    ) -> List[str]:
        if not config:
            config = Config()

        parse_text = _parse_yaml if config.is_yaml else _parse_json
        try:
            parsed = _parse_to_dict(json_str_or_dict_or_file_path, parse_text)
        except ParseError as err:
            return [f"Problem reading/parsing {config.input_type}: {err.root_cause}"]

        if not isinstance(parsed, dict):
            return []

        problems = [
            p
            for p in StateMachine(StateFactory(), parsed).validate()
            if p.type not in config.ignored_problem_types
        ]

        return [str(p) for p in problems]


def _parse_to_dict(
    json_str_or_dict_or_file_path: Union[str, Any],
    parse_text: Callable[[str], Dict[str, Any]],
) -> Any:
    if isinstance(json_str_or_dict_or_file_path, dict):
        return json_str_or_dict_or_file_path

    if os.path.isfile(json_str_or_dict_or_file_path):
        with open(json_str_or_dict_or_file_path) as file:
            text = file.read()
    else:
        text = json_str_or_dict_or_file_path

    return parse_text(text)


def _parse_json(text: str) -> Any:
    try:
        return json.loads(text)
    except json.JSONDecodeError as err:
        raise ParseError(err)


def _parse_yaml(text: str) -> Any:
    import yaml

    return yaml.safe_load(text)



import pytest

from statelint.linter import Linter


def test_null_type():
    state_machine = {"StartAt": "x", "States": {"x": {"Type": None, "End": True}}}
    assert Linter.validate(state_machine) == [
        "State Machine.States.x.Type should be non-null",
        'State Machine.States.x.Type is "", not one of the allowed values '
        '["Pass", "Succeed", "Fail", "Task", "Choice", "Wait", "Parallel", "Map"]',
        'Field "End" not allowed in State Machine.States.x',
    ]


def test_illegal_type():
    state_machine = {"StartAt": "x", "States": {"x": {"Type": "?", "End": True}}}
    assert Linter.validate(state_machine) == [
        'State Machine.States.x.Type is "?", not one of the allowed values '
        '["Pass", "Succeed", "Fail", "Task", "Choice", "Wait", "Parallel", "Map"]',
        'Field "End" not allowed in State Machine.States.x',
    ]


def test_null_resource():
    state_machine = {
        "StartAt": "x",
        "States": {"x": {"Type": "Task", "Resource": 1, "End": True}},
    }
    assert Linter.validate(state_machine) == [
        "State Machine.States.x.Resource is 1 but should be A URI"
    ]


@pytest.mark.parametrize(
    "seconds,problems",
    [
        (0, []),
        (-1, ["State Machine.States.x.Seconds is -1 but allowed floor is 0"]),
        (1, []),
    ],
)
def test_zero_seconds(seconds, problems):
    state_machine = {
        "StartAt": "x",
        "States": {"x": {"Type": "Wait", "Seconds": seconds, "End": True}},
    }
    assert Linter.validate(state_machine) == problems



import pytest

from statelint.linter import Linter


@pytest.mark.parametrize(
    "comparison,comparison_value",
    [
        ("StringEquals", "x"),
        ("StringLessThan", "x"),
        ("StringGreaterThan", "x"),
        ("StringLessThanEquals", "x"),
        ("StringGreaterThanEquals", "x"),
        ("NumericEquals", 1.2),
        ("NumericLessThan", -1),
        ("NumericGreaterThan", 0),
        ("NumericLessThanEquals", 0.0),
        ("NumericGreaterThanEquals", 999999999),
        ("BooleanEquals", False),
        ("TimestampEquals", "2001-01-01T12:00:00Z"),
        ("TimestampLessThan", "2001-01-01T12:00:00Z"),
        ("TimestampGreaterThan", "2001-01-01T12:00:00Z"),
        ("TimestampLessThanEquals", "2001-01-01T12:00:00Z"),
        ("TimestampGreaterThanEquals", "2001-01-01T12:00:00Z"),
        ("IsNull", True),
        ("IsPresent", True),
        ("IsNumeric", True),
        ("IsString", True),
        ("IsBoolean", True),
        ("IsTimestamp", True),
        ("StringMatches", "a*"),
        ("StringEqualsPath", "$"),
        ("StringLessThanPath", "$.x"),
        ("StringGreaterThanPath", "$"),
        ("StringLessThanEqualsPath", "$"),
        ("StringGreaterThanEqualsPath", "$"),
        ("NumericEqualsPath", "$"),
        ("NumericLessThanPath", "$"),
        ("NumericGreaterThanPath", "$"),
        ("NumericLessThanEqualsPath", "$"),
        ("NumericGreaterThanEqualsPath", "$"),
        ("BooleanEqualsPath", "$"),
        ("TimestampEqualsPath", "$"),
        ("TimestampLessThanPath", "$"),
        ("TimestampGreaterThanPath", "$"),
        ("TimestampLessThanEqualsPath", "$"),
        ("TimestampGreaterThanEqualsPath", "$"),
    ],
)
def test_ok(comparison, comparison_value):
    state_machine = {
        "States": {
            "x": {
                "Type": "Choice",
                "Comment": "aaa",
                "Choices": [
                    {
                        "Comment": "aaa",
                        "And": [
                            {
                                "Or": [
                                    {
                                        "Variable": "$",
                                        comparison: comparison_value,
                                        "Comment": "aaa",
                                    },
                                    {"Variable": "$", "IsNull": False},
                                ]
                            },
                            {
                                "Not": {
                                    "And": [
                                        {"Variable": "$", "IsNull": False},
                                        {"Variable": "$", "IsNull": True},
                                    ]
                                }
                            },
                        ],
                        "Next": "y",
                    }
                ],
                "Default": "z",
            },
            "y": {"Type": "Succeed"},
            "z": {"Type": "Succeed"},
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == []


@pytest.mark.parametrize(
    "comparison,comparison_value,problems",
    [
        (
            "StringEquals",
            None,
            [
                "State Machine.States.x.Choices[0].And[0].Or[0].StringEquals "
                "should be non-null"
            ],
        ),
        (
            "StringLessThan",
            1,
            [
                "State Machine.States.x.Choices[0].And[0].Or[0].StringLessThan "
                "is 1 but should be a String"
            ],
        ),
        (
            "NumericEquals",
            None,
            [
                "State Machine.States.x.Choices[0].And[0].Or[0].NumericEquals "
                "should be non-null"
            ],
        ),
        (
            "NumericLessThan",
            "1",
            [
                "State Machine.States.x.Choices[0].And[0].Or[0].NumericLessThan "
                'is "1" but should be numeric'
            ],
        ),
        (
            "BooleanEquals",
            None,
            [
                "State Machine.States.x.Choices[0].And[0].Or[0].BooleanEquals "
                "should be non-null"
            ],
        ),
        (
            "TimestampEquals",
            None,
            [
                "State Machine.States.x.Choices[0].And[0].Or[0].TimestampEquals "
                "should be non-null"
            ],
        ),
        (
            "TimestampLessThan",
            "x",
            [
                "State Machine.States.x.Choices[0].And[0].Or[0].TimestampLessThan "
                'is "x" but should be an RFC3339 timestamp'
            ],
        ),
        (
            "IsNull",
            None,
            [
                "State Machine.States.x.Choices[0].And[0].Or[0].IsNull should be "
                "non-null"
            ],
        ),
        (
            "IsPresent",
            1,
            [
                "State Machine.States.x.Choices[0].And[0].Or[0].IsPresent is 1 "
                "but should be a Boolean"
            ],
        ),
        (
            "IsNumeric",
            "True",
            [
                "State Machine.States.x.Choices[0].And[0].Or[0].IsNumeric is "
                '"True" but should be a Boolean'
            ],
        ),
        (
            "StringMatches",
            None,
            [
                "State Machine.States.x.Choices[0].And[0].Or[0].StringMatches "
                "should be non-null"
            ],
        ),
        (
            "StringEqualsPath",
            None,
            [
                "State Machine.States.x.Choices[0].And[0].Or[0].StringEqualsPath "
                "should be non-null"
            ],
        ),
        (
            "StringLessThanPath",
            "x",
            [
                "State Machine.States.x.Choices[0].And[0].Or[0].StringLessThanPath is "
                '"x" but should be a Reference Path'
            ],
        ),
    ],
)
def test_ng(comparison, comparison_value, problems):
    state_machine = {
        "States": {
            "x": {
                "Type": "Choice",
                "Choices": [
                    {
                        "And": [
                            {"Or": [{"Variable": "$", comparison: comparison_value}]}
                        ],
                        "Next": "y",
                    }
                ],
            },
            "y": {"Type": "Succeed"},
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == problems


def test_ok_all_of_or_not_and():
    state_machine = {
        "States": {
            "x": {
                "Type": "Choice",
                "Choices": [
                    {
                        "And": [{"Variable": "$", "IsNull": True}],
                        "Or": [{"Variable": "$", "IsNull": False}],
                        "Not": {"Variable": "$", "IsNull": False},
                        "Next": "y",
                    }
                ],
                "Default": "z",
            },
            "y": {"Type": "Succeed"},
            "z": {"Type": "Succeed"},
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == []


def test_empty_at_deep_choice():
    state_machine = {
        "States": {
            "x": {
                "Type": "Choice",
                "Choices": [
                    {
                        "And": [
                            {
                                "Or": [
                                    {"Variable": "$", "IsNull": True, "Not": {}},
                                    {"Variable": "$", "IsNull": False},
                                ]
                            },
                            {"Not": {"And": []}},
                        ],
                        "Next": "y",
                    }
                ],
                "Default": "z",
            },
            "y": {"Type": "Succeed"},
            "z": {"Type": "Succeed"},
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        'State Machine.States.x.Choices[0].And[0].Or[0] has forbidden field "Not"',
        "State Machine.States.x.Choices[0].And[1].Not.And is empty, non-empty required",
    ]


def test_next_in_nested_rule():
    state_machine = {
        "States": {
            "x": {
                "Type": "Choice",
                "Choices": [
                    {
                        "And": [
                            {"Or": [{"Variable": "$", "IsNull": True, "Next": "y"}]},
                        ],
                        "Next": "y",
                    }
                ],
                "Default": "y",
            },
            "y": {"Type": "Succeed"},
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        'State Machine.States.x.Choices[0].And[0].Or[0] has forbidden field "Next"'
    ]



import itertools
from typing import Any, Dict, List

from ..fields import BRANCHES, Field
from ..problem import Problem
from .container_state import ContainerState
from .factory import NodeFactory
from .mixins import (
    CatchMixin,
    NextXorEndMixin,
    ParametersMixin,
    ResultPathMixin,
    ResultSelectorMixin,
    RetryMixin,
    TimeoutSecondsMixin,
)
from .node import NameAndPath, StatePath
from .state import State


class ParallelState(
    NextXorEndMixin,
    ResultPathMixin,
    ParametersMixin,
    ResultSelectorMixin,
    TimeoutSecondsMixin,
    CatchMixin,
    RetryMixin,
    State,
):
    def __init__(
        self, node_factory: NodeFactory, state_path: StatePath, state: Dict[str, Any]
    ) -> None:
        super().__init__(state_path, state)
        self._node_factory = node_factory
        self._branches = self._get_branches(state)

    def _get_branches(self, state: Dict[str, Any]) -> List[ContainerState]:
        branches = state.get(BRANCHES.name)
        if not isinstance(branches, list):
            return []
        return [
            ContainerState(
                self._node_factory,
                self.state_path.make_child(BRANCHES, idx),
                element,
            )
            for idx, element in enumerate(state[BRANCHES.name])
            if isinstance(element, dict)
        ]

    @property
    def required_fields(self) -> List[Field]:
        return [*super().required_fields, BRANCHES]

    def get_children(self) -> List[NameAndPath]:
        return list(
            itertools.chain.from_iterable(b.get_children() for b in self._branches)
        )

    def validate(self) -> List[Problem]:
        return super().validate() + [
            p for branch in self._branches for p in branch.validate()
        ]



from typing import Any, List

from ..problem import ProblemPredicate, ProblemType
from .base import Field, NonNullMixin


class NullableStrField(Field):
    @property
    def problem_type(self) -> ProblemType:
        return ProblemType.STRING

    def validate(self, value: Any) -> List[ProblemPredicate]:
        if value is None:
            return []
        return self.check_type(value, str, self.problem_type)


class StrField(NonNullMixin, NullableStrField):
    pass


class EnumStrField(StrField):
    def __init__(self, name: str, choices: List[str]):
        super().__init__(name)
        self.choices = choices

    def validate(self, value: Any) -> List[ProblemPredicate]:
        problems = super().validate(value)
        str_val = "" if value is None else str(value)
        if str_val not in self.choices:
            choices = ", ".join(f'"{c}"' for c in self.choices)
            problems += [
                ProblemPredicate(
                    f' is "{str_val}", not one of the allowed values [{choices}]'
                )
            ]

        return problems



from abc import ABC
from typing import Any, List, Set, Tuple, Union

from ..problem import ProblemPredicate, ProblemType
from .common import to_json


class Field(ABC):
    name: str

    def __init__(self, name: str) -> None:
        self.name = name

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.name})"

    def validate(self, value: Any) -> List[ProblemPredicate]:
        return []

    @staticmethod
    def check_type(
        value: Any, _type: Union[type, Tuple[type, ...]], problem_type: ProblemType
    ) -> List[ProblemPredicate]:
        if isinstance(value, _type):
            return []
        return [
            ProblemPredicate(
                f" is {to_json(value)} but should be {problem_type}", problem_type
            )
        ]

    def get_fields(self) -> List["Field"]:
        return [self]

    def validate_as_required(self, names: Set[str]) -> List[ProblemPredicate]:
        if self.name in names:
            return []
        return [ProblemPredicate(f' does not have required field "{self.name}"')]

    def validate_as_optional(self, names: Set[str]) -> List[ProblemPredicate]:
        return []


class NonNullMixin(Field):
    def validate(self, value: Any) -> List[ProblemPredicate]:
        problems = super().validate(value)
        if problems:
            return problems
        if value is None:
            return [ProblemPredicate(" should be non-null")]
        return []



from typing import Any, List

from ..problem import ProblemPredicate, ProblemType
from .base import Field, NonNullMixin


class ObjectField(NonNullMixin, Field):
    def validate(self, value: Any) -> List[ProblemPredicate]:
        return self.check_type(value, dict, ProblemType.OBJECT)



from typing import List

from ..fields import (
    CREDENTIALS,
    HEARTBEAT_SECONDS,
    HEARTBEAT_SECONDS_PATH,
    RESOURCE,
    TIMEOUT_SECONDS,
    TIMEOUT_SECONDS_PATH,
    Field,
    OneOfField,
)
from .mixins import (
    CatchMixin,
    NextXorEndMixin,
    ParametersMixin,
    ResultPathMixin,
    ResultSelectorMixin,
    RetryMixin,
    TimeoutSecondsMixin,
)
from .state import State


class TaskState(
    NextXorEndMixin,
    ResultPathMixin,
    ParametersMixin,
    ResultSelectorMixin,
    TimeoutSecondsMixin,
    CatchMixin,
    RetryMixin,
    State,
):
    @property
    def optional_fields(self) -> List[Field]:
        return [
            *super().optional_fields,
            RESOURCE,
            OneOfField(TIMEOUT_SECONDS, TIMEOUT_SECONDS_PATH),
            OneOfField(HEARTBEAT_SECONDS, HEARTBEAT_SECONDS_PATH),
            CREDENTIALS,
        ]



from typing import List

from ...fields import END, NEXT, RESULT, TIMEOUT_SECONDS, Field
from ..node import Node
from .catcher_mixin import CatchMixin
from .choices_mixin import ChoicesMixin
from .payload_template_mixin import (
    BatchInputMixin,
    ItemSelectorMixin,
    ParametersMixin,
    ResultSelectorMixin,
)
from .result_path_mixin import ResultPathMixin
from .retrier_mixin import RetryMixin


class ResultMixin(Node):
    @property
    def optional_fields(self) -> List[Field]:
        return [*super().optional_fields, RESULT]


class TimeoutSecondsMixin(Node):
    @property
    def optional_fields(self) -> List[Field]:
        return [*super().optional_fields, TIMEOUT_SECONDS]


class NextXorEndMixin(Node):
    @property
    def required_fields(self) -> List[Field]:
        required_field = END if END.name in self._state else NEXT
        return [*super().required_fields, required_field]

    @property
    def optional_fields(self) -> List[Field]:
        return [*super().optional_fields, NEXT, END]

    @property
    def forbidden_fields(self) -> List[Field]:
        if END.name in self._state:
            return [NEXT]
        return [END]


class NextMixin(Node):
    @property
    def optional_fields(self) -> List[Field]:
        return [*super().optional_fields, NEXT]


class EndMixin(Node):
    @property
    def forbidden_fields(self) -> List[Field]:
        return [NEXT]



from typing import List

from ..fields import SECONDS, SECONDS_PATH, TIMESTAMP, TIMESTAMP_PATH, Field, OneOfField
from .mixins import NextXorEndMixin
from .state import State


class WaitState(NextXorEndMixin, State):
    @property
    def required_fields(self) -> List[Field]:
        return [
            *super().required_fields,
            OneOfField(SECONDS, SECONDS_PATH, TIMESTAMP, TIMESTAMP_PATH),
        ]

    @property
    def optional_fields(self) -> List[Field]:
        return [*super().optional_fields]



from typing import Any, List

from ..problem import ProblemPredicate, ProblemType
from .base import Field, NonNullMixin


class BoolField(NonNullMixin, Field):
    def validate(self, value: Any) -> List[ProblemPredicate]:
        problems = super().validate(value)
        if problems:
            return problems
        return self.check_type(value, bool, ProblemType.BOOLEAN)



import pytest

from statelint.linter import Linter

from .common import get_path


def test_invalid_json_str():
    assert Linter.validate("a") == [
        "Problem reading/parsing JSON: Expecting value: line 1 column 1 (char 0)"
    ]


def test_invalid_json_file():
    file_path = get_path("invalid.json")
    assert Linter.validate(file_path) == [
        "Problem reading/parsing JSON: Expecting ',' delimiter: line 2 "
        "column 11 (char 12)"
    ]


def test_null():
    assert Linter.validate("null") == []


@pytest.mark.parametrize(
    "filepath",
    [
        ("states-uuid-invocation.json"),
    ],
)
def test_valid_json(filepath):
    assert Linter().validate(get_path(filepath)) == []



import pytest

from statelint.linter import Linter


@pytest.mark.parametrize(
    "parameters,expected",
    [
        (None, []),
        ([], []),
        ("x", []),
        (
            {
                "a": {
                    "b.$": "$[0].c",
                    "d.$": "States.Format('{}_{}', $$.Execution.Name, $.e)",
                }
            },
            [],
        ),
        (
            {
                "a": {
                    "b.$": "States.Format('{}_{}', $$.Execution.Name, $.e",
                }
            },
            [
                'Field "b.$" of Parameters at "State Machine.States.x.a" is not '
                "a JSONPath or intrinsic function expression"
            ],
        ),
        (
            {"a": [{"b.$": "c", "d.$": "e"}]},
            [
                # TODO: awslabs' output doesn't contain `.States`.
                'Field "b.$" of Parameters at "State Machine.States.x.a[0]" is not '
                "a JSONPath or intrinsic function expression",
                'Field "d.$" of Parameters at "State Machine.States.x.a[0]" is not '
                "a JSONPath or intrinsic function expression",
            ],
        ),
        (
            {"a": [{"b": [{}, {"c": {"d.$": "x"}}]}]},
            [
                'Field "d.$" of Parameters at "State Machine.States.x.a[0].b[1].c" '
                "is not a JSONPath or intrinsic function expression",
            ],
        ),
        (
            {"a.$": None},
            [
                'Field "a.$" of Parameters at "State Machine.States.x" is not '
                "a JSONPath or intrinsic function expression",
            ],
        ),
    ],
)
def test_main(parameters, expected):
    state_machine = {
        "States": {"x": {"Type": "Pass", "End": True, "Parameters": parameters}},
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == expected



import pytest

from statelint.fields.pattern_field import JsonPathField, RefPathField


@pytest.mark.parametrize(
    "text,is_ok",
    [
        # should allow default paths
        ("$", True),
        # should reject obvious botches
        ("x", False),
        (".x", False),
        ("x.y.z", False),
        ("$.~.bar", False),
        # should allow reference paths
        ("$.foo.bar", True),
        ("$..x", True),
        ("$.foo.bar.baz.biff..blecch", True),
        ("$.café_au_lait", True),
        ("$['foo']['bar']", True),
        ("$['foo']['bar']['baz']['biff']..blecch", True),
        ("$['café_au_lait']", True),
        ("$..author", True),
        ("$..book[2]", True),
        ("$['豆柴']", True),
        # should distinguish between non-paths, paths, and reference paths
        ("$.store.book[*].author", False),
        ("$..author", True),
        ("$.store.*", False),
        ("$..book[2]", True),
        ("$..book[0,1]", False),
        ("$..book[:2]", False),
        ("$..book[1:2]", False),
        ("$..book[-2:]", False),
        ("$..book[2:]", False),
        ("$..*", False),
    ],
)
def test_ref_field(text, is_ok):
    field = RefPathField("test")
    assert (not field.validate(text)) == is_ok


@pytest.mark.parametrize(
    "text,is_ok",
    [
        # should allow default paths
        ("$", True),
        # should do simple paths
        ("$.foo.bar", True),
        ("$..x", True),
        ("$.foo.bar.baz.biff..blecch", True),
        ("$.café_au_lait", True),
        ("$['foo']", True),
        ("$[3]", True),
        # should reject obvious botches
        ("x", False),
        (".x", False),
        ("x.y.z", False),
        ("$.~.bar", False),
        # should accept paths with bracket notation
        ("$['foo']['bar']", True),
        ("$['foo']['bar']['baz']['biff']..blecch", True),
        ("$['café_au_lait']", True),
        # should accept some Jayway JsonPath examples
        ("$.store.book[*].author", True),
        ("$..author", True),
        ("$.store.*", True),
        ("$..book[2]", True),
        ("$..book[0,1]", True),
        ("$..book[:2]", True),
        ("$..book[1:2]", True),
        ("$..book[-2:]", True),
        ("$..book[2:]", True),
        ("$..*", True),
        # should distinguish between non-paths, paths, and reference paths
        ("$.store.book[*].author", True),
        ("$..author", True),
        ("$.store.*", True),
        ("$..book[2]", True),
        ("$..book[0,1]", True),
        ("$..book[:2]", True),
        ("$..book[1:2]", True),
        ("$..book[-2:]", True),
        ("$..book[2:]", True),
        ("$..*", True),
    ],
)
def test_main(text, is_ok):
    field = JsonPathField("test")
    assert (not field.validate(text)) == is_ok


def test_message():
    field = JsonPathField("test")
    assert [p.predicate for p in field.validate("x")] == [
        ' is "x" but should be a JSONPath'
    ]


def test_repr():
    assert repr(JsonPathField("x")) == "JsonPathField(x)"



from typing import Any, Dict, List

from ..fields import VERSION, Field
from .container_state import ContainerState
from .factory import NodeFactory
from .mixins import TimeoutSecondsMixin
from .node import StatePath


class StateMachine(TimeoutSecondsMixin, ContainerState):
    def __init__(self, node_factory: NodeFactory, state: Dict[str, Any]) -> None:
        super().__init__(node_factory, StatePath("State Machine"), state)

    @property
    def optional_fields(self) -> List[Field]:
        return [*super().optional_fields, VERSION]



from typing import Any, List

from ...fields import CATCH, ERROR_EQUALS, NEXT, Field
from ...problem import Problem
from ..node import NameAndPath, Node
from .result_path_mixin import ResultPathMixin


class Catcher(ResultPathMixin, Node):
    @property
    def required_fields(self) -> List[Field]:
        return [*super().required_fields, ERROR_EQUALS, NEXT]


class CatchMixin(Node):
    @property
    def optional_fields(self) -> List[Field]:
        return [*super().optional_fields, CATCH]

    def validate(self) -> List[Problem]:
        problems = super().validate()
        catchers = self._state.get(CATCH.name)
        if not isinstance(catchers, list):
            return problems
        return (
            problems
            + [
                p
                for idx, element in enumerate(catchers)
                if isinstance(element, dict)
                for p in Catcher(
                    self.state_path.make_child(CATCH, idx), element
                ).validate()
            ]
            + self._validate_error_equals(catchers)
        )

    def _validate_error_equals(self, catchers: List[Any]) -> List[Problem]:
        problems = []
        for idx, catcher in enumerate(catchers):
            if not isinstance(catcher, dict):
                continue
            error_equals = catcher.get(ERROR_EQUALS.name)
            if not isinstance(error_equals, list):
                continue

            if "States.ALL" in error_equals:
                if idx != len(catchers) - 1 or len(error_equals) != 1:
                    state_path = self.state_path.make_child(CATCH, idx)
                    problems.append(
                        Problem(
                            f"{state_path}: States.ALL can only "
                            "appear in the last element, and by itself."
                        ),
                    )

        return problems

    def get_reachable_states(self) -> List[NameAndPath]:
        states = super().get_reachable_states()
        state = self._state
        if not isinstance(state.get(CATCH.name), list):
            return states

        return states + [
            NameAndPath(element[NEXT.name], self.state_path.make_child(NEXT))
            for element in state[CATCH.name]
            if isinstance(element, dict) and isinstance(element.get(NEXT.name), str)
        ]



from abc import ABC, abstractmethod
from typing import Any, Optional

from ..node import Node, StatePath


class NodeFactory(ABC):
    @abstractmethod
    def get(self, state_path: StatePath, state: Any) -> Optional[Node]:
        pass  # pragma: no cover



from statelint.linter import Linter


def test_ok_next():
    state_machine = {
        "States": {"x": {"Type": "Pass", "Next": "y"}, "y": {"Type": "Fail"}},
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == []


def test_ok_end():
    state_machine = {
        "States": {
            "x": {"Type": "Pass", "Parameters": {}, "Result": True, "End": True}
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == []


def test_ok_result_path():
    state_machine = {
        "States": {"x": {"Type": "Pass", "End": True, "ResultPath": "$"}},
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == []


def test_both_next_and_end():
    state_machine = {
        "States": {"x": {"Type": "Pass", "Next": "y", "End": True}},
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        'State Machine.States.x has forbidden field "Next"',
        'No state found named "y", referenced at State Machine.States.x.Next',
    ]


def test_empty():
    state_machine = {
        "States": {"x": {"Type": "Pass"}},
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        'State Machine.States.x does not have required field "Next"',
        "No terminal state found in machine at State Machine.States",
    ]


def test_result_path():
    state_machine = {
        "States": {"x": {"Type": "Pass", "End": True, "ResultPath": "x"}},
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        'State Machine.States.x.ResultPath is "x" but should be a Reference Path'
    ]



from collections import OrderedDict
from typing import List, Set

from ..problem import ProblemPredicate
from .any_field import AnyField
from .base import Field
from .common import to_json


class OneOfField(AnyField):
    def __init__(self, *fields: Field) -> None:
        # use OrderedDict as set to preserve order
        self._fields = OrderedDict((f.name, f) for f in fields)
        fields_str = ", ".join(to_json(f) for f in self._fields)
        super().__init__(f"[{fields_str}]")

    def get_fields(self) -> List["Field"]:
        return [*super().get_fields(), *self._fields.values()]

    def validate_as_required(self, names: Set[str]) -> List[ProblemPredicate]:
        intersection_count = len(names & set(self._fields))
        if intersection_count == 0:
            return [ProblemPredicate(f" does not have required field from {self.name}")]
        return []

    def validate_as_optional(self, names: Set[str]) -> List[ProblemPredicate]:
        intersection_count = len(names & set(self._fields))
        if intersection_count > 1:
            return [ProblemPredicate(f" may have only one of {self.name}")]
        return []



from statelint.linter import Linter


def test_ok():
    state_machine = {
        "States": {"x": {"Type": "Wait", "End": True, "Seconds": 1}},
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == []


def test_empty():
    state_machine = {
        "States": {"x": {"Type": "Wait", "End": True}},
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        "State Machine.States.x does not have required field from "
        '["Seconds", "SecondsPath", "Timestamp", "TimestampPath"]',
    ]


def test_null_seconds_path():
    state_machine = {
        "States": {"x": {"Type": "Wait", "SecondsPath": None, "End": True}},
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        "State Machine.States.x.SecondsPath should be non-null",
    ]


def test_2_out_of_4():
    state_machine = {
        "States": {
            "x": {"Type": "Wait", "End": True, "Seconds": 1, "TimestampPath": "$"}
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        "State Machine.States.x may have only one of "
        '["Seconds", "SecondsPath", "Timestamp", "TimestampPath"]'
    ]


def test_bad_timestamp():
    state_machine = {
        "States": {"x": {"Type": "Wait", "End": True, "Timestamp": 1}},
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        "State Machine.States.x.Timestamp is 1 but should be an RFC3339 timestamp"
    ]


def test_null_timestamp():
    state_machine = {
        "States": {"x": {"Type": "Wait", "End": True, "Timestamp": None}},
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        "State Machine.States.x.Timestamp should be non-null"
    ]



from typing import List

from ..fields import INPUT_PATH, OUTPUT_PATH, TYPE, Field
from .node import Node


class State(Node):
    @property
    def required_fields(self) -> List[Field]:
        return [*super().required_fields, TYPE]

    @property
    def optional_fields(self) -> List[Field]:
        return [*super().optional_fields, INPUT_PATH, OUTPUT_PATH]



from typing import List

from ..fields import (
    CAUSE,
    CAUSE_PATH,
    ERROR,
    ERROR_PATH,
    INPUT_PATH,
    OUTPUT_PATH,
    Field,
    OneOfField,
)
from .mixins import EndMixin
from .state import State


class FailState(EndMixin, State):
    @property
    def optional_fields(self) -> List[Field]:
        return [
            *super().optional_fields,
            OneOfField(CAUSE, CAUSE_PATH),
            OneOfField(ERROR, ERROR_PATH),
        ]

    @property
    def forbidden_fields(self) -> List[Field]:
        return [INPUT_PATH, OUTPUT_PATH, *super().forbidden_fields]



from abc import ABC
from collections import OrderedDict
from typing import Any, List

from ...fields import (
    AND,
    BOOLEAN_EQUALS,
    BOOLEAN_EQUALS_PATH,
    CHOICES,
    DEFAULT,
    IS_BOOLEAN,
    IS_NULL,
    IS_NUMERIC,
    IS_PRESENT,
    IS_STRING,
    IS_TIMESTAMP,
    NEXT,
    NOT,
    NUMERIC_EQUALS,
    NUMERIC_EQUALS_PATH,
    NUMERIC_GREATER_THAN,
    NUMERIC_GREATER_THAN_EQUALS,
    NUMERIC_GREATER_THAN_EQUALS_PATH,
    NUMERIC_GREATER_THAN_PATH,
    NUMERIC_LESS_THAN,
    NUMERIC_LESS_THAN_EQUALS,
    NUMERIC_LESS_THAN_EQUALS_PATH,
    NUMERIC_LESS_THAN_PATH,
    OR,
    STRING_EQUALS,
    STRING_EQUALS_PATH,
    STRING_GREATER_THAN,
    STRING_GREATER_THAN_EQUALS,
    STRING_GREATER_THAN_EQUALS_PATH,
    STRING_GREATER_THAN_PATH,
    STRING_LESS_THAN,
    STRING_LESS_THAN_EQUALS,
    STRING_LESS_THAN_EQUALS_PATH,
    STRING_LESS_THAN_PATH,
    STRING_MATCHES,
    TIMESTAMP_EQUALS,
    TIMESTAMP_EQUALS_PATH,
    TIMESTAMP_GREATER_THAN,
    TIMESTAMP_GREATER_THAN_EQUALS,
    TIMESTAMP_GREATER_THAN_EQUALS_PATH,
    TIMESTAMP_GREATER_THAN_PATH,
    TIMESTAMP_LESS_THAN,
    TIMESTAMP_LESS_THAN_EQUALS,
    TIMESTAMP_LESS_THAN_EQUALS_PATH,
    TIMESTAMP_LESS_THAN_PATH,
    VARIABLE,
    Field,
    OneOfField,
)
from ...problem import Problem
from ..node import NameAndPath, Node

COMPARISONS = OneOfField(
    STRING_EQUALS,
    STRING_LESS_THAN,
    STRING_GREATER_THAN,
    STRING_LESS_THAN_EQUALS,
    STRING_GREATER_THAN_EQUALS,
    NUMERIC_EQUALS,
    NUMERIC_LESS_THAN,
    NUMERIC_GREATER_THAN,
    NUMERIC_LESS_THAN_EQUALS,
    NUMERIC_GREATER_THAN_EQUALS,
    BOOLEAN_EQUALS,
    TIMESTAMP_EQUALS,
    TIMESTAMP_LESS_THAN,
    TIMESTAMP_GREATER_THAN,
    TIMESTAMP_LESS_THAN_EQUALS,
    TIMESTAMP_GREATER_THAN_EQUALS,
    STRING_EQUALS_PATH,
    STRING_LESS_THAN_PATH,
    STRING_GREATER_THAN_PATH,
    STRING_LESS_THAN_EQUALS_PATH,
    STRING_GREATER_THAN_EQUALS_PATH,
    NUMERIC_EQUALS_PATH,
    NUMERIC_LESS_THAN_PATH,
    NUMERIC_GREATER_THAN_PATH,
    NUMERIC_LESS_THAN_EQUALS_PATH,
    NUMERIC_GREATER_THAN_EQUALS_PATH,
    BOOLEAN_EQUALS_PATH,
    TIMESTAMP_EQUALS_PATH,
    TIMESTAMP_LESS_THAN_PATH,
    TIMESTAMP_GREATER_THAN_PATH,
    TIMESTAMP_LESS_THAN_EQUALS_PATH,
    TIMESTAMP_GREATER_THAN_EQUALS_PATH,
    IS_NULL,
    IS_PRESENT,
    IS_NUMERIC,
    IS_STRING,
    IS_BOOLEAN,
    IS_TIMESTAMP,
    STRING_MATCHES,
)


class BaseChoiceRule(Node, ABC):
    @property
    def optional_fields(self) -> List[Field]:
        fields = super().optional_fields
        if VARIABLE.name in self._state:
            return [*fields, VARIABLE, COMPARISONS]
        return [*fields, AND, OR, NOT]

    @property
    def forbidden_fields(self) -> List[Field]:
        fields = [f for f in super().forbidden_fields]
        if VARIABLE.name in self._state:
            return [*fields, AND, OR, NOT]
        return fields

    def validate(self) -> List[Problem]:
        problems = list(super().validate())
        for field in (AND.name, OR.name):
            if isinstance(self._state.get(field), list):
                for idx, child in enumerate(self._state[field]):
                    problems.extend(self._validate_child(f"{field}[{idx}]", child))

        problems.extend(self._validate_child(NOT.name, self._state.get(NOT.name)))

        return problems

    def _validate_child(self, field_name: str, child: Any) -> List[Problem]:
        if isinstance(child, dict):
            return NestedChoiceRule(
                self.state_path.make_child(field_name), child
            ).validate()
        return []


class ChoiceRule(BaseChoiceRule):
    @property
    def required_fields(self) -> List[Field]:
        return [*super().required_fields, NEXT]


class NestedChoiceRule(BaseChoiceRule):
    @property
    def forbidden_fields(self) -> List[Field]:
        return [NEXT, *super().forbidden_fields]


class ChoicesMixin(Node):
    @property
    def required_fields(self) -> List[Field]:
        return [*super().required_fields, CHOICES]

    def validate(self) -> List[Problem]:
        problems = super().validate()
        if not isinstance(self._state.get(CHOICES.name), list):
            return problems
        return problems + [
            p
            for idx, element in enumerate(self._state[CHOICES.name])
            if isinstance(element, dict)
            for p in ChoiceRule(
                self.state_path.make_child(CHOICES.name, idx), element
            ).validate()
        ]

    def get_reachable_states(self) -> List[NameAndPath]:
        # use OrderedDict as set to preserve order
        reachable_states = OrderedDict({s: 0 for s in super().get_reachable_states()})
        if isinstance(self._state.get(DEFAULT.name), str):
            name = self._state[DEFAULT.name]
            reachable_states[NameAndPath(name, self.state_path.make_child(DEFAULT))] = 0
        if isinstance(self._state.get(CHOICES.name), list):
            for idx, choice in enumerate(self._state[CHOICES.name]):
                if isinstance(choice, dict) and isinstance(choice.get(NEXT.name), str):
                    state_path = self.state_path.make_child(CHOICES, idx, NEXT)
                    reachable_states[NameAndPath(choice[NEXT.name], state_path)] = 0
        return list(reachable_states)



from .problem_type import ProblemType


class ProblemPredicate:
    predicate: str
    type: ProblemType

    def __init__(
        self, predicate: str, _type: ProblemType = ProblemType.UNKNOWN
    ) -> None:
        self.predicate = predicate
        self.type = _type

    def __str__(self) -> str:
        return self.predicate



from typing import Any, List, Type

from ..problem import ProblemPredicate, ProblemType
from .base import Field, NonNullMixin


class ListField(NonNullMixin, Field):
    def __init__(self, name: str, element_field_type: Type[Field]) -> None:
        super().__init__(name)
        self._element_field_type = element_field_type(name)

    def validate(self, value: Any) -> List[ProblemPredicate]:
        problems = super().validate(value)
        if problems:
            return problems
        problems = self.check_type(value, list, ProblemType.ARRAY)
        if problems:
            return problems
        return [
            ProblemPredicate(f"[{idx}]{problem}")
            for idx, element in enumerate(value)
            for problem in self._element_field_type.validate(element)
        ]


class NonEmptyListField(ListField):
    def validate(self, value: Any) -> List[ProblemPredicate]:
        problems = super().validate(value)
        if problems:
            return problems
        if len(value) == 0:
            return [ProblemPredicate(" is empty, non-empty required")]
        return []



from typing import List

from ...fields import (
    BACKOFF_RATE,
    ERROR_EQUALS,
    INTERVAL_SECONDS,
    JITTER_STRATEGY,
    MAX_ATTEMPTS,
    MAX_DELAY_SECONDS,
    RETRY,
    Field,
)
from ...problem import Problem
from ..node import Node
from .result_path_mixin import ResultPathMixin


class Retrier(ResultPathMixin, Node):
    @property
    def required_fields(self) -> List[Field]:
        return [*super().required_fields, ERROR_EQUALS]

    @property
    def optional_fields(self) -> List[Field]:
        return [
            *super().optional_fields,
            INTERVAL_SECONDS,
            MAX_ATTEMPTS,
            BACKOFF_RATE,
            MAX_DELAY_SECONDS,
            JITTER_STRATEGY,
        ]


class RetryMixin(Node):
    @property
    def optional_fields(self) -> List[Field]:
        return [*super().optional_fields, RETRY]

    def validate(self) -> List[Problem]:
        state = self._state
        problems = super().validate()
        if not isinstance(state.get(RETRY.name), list):
            return problems
        return problems + [
            p
            for idx, element in enumerate(state[RETRY.name])
            if isinstance(element, dict)
            for p in Retrier(self.state_path.make_child(RETRY, idx), element).validate()
        ]



from .mixins import NextXorEndMixin, ParametersMixin, ResultMixin, ResultPathMixin
from .state import State


class PassState(ResultPathMixin, ResultMixin, ParametersMixin, NextXorEndMixin, State):
    pass



import pytest

from statelint.linter import Linter


def test_ok():
    state_machine = {
        "States": {
            "x": {
                "Type": "Map",
                "ResultPath": None,
                "Parameters": {},
                "ResultSelector": {},
                "ItemsPath": "$$.x",
                "MaxConcurrency": 9999999999,
                "Retry": [{"ErrorEquals": ["E"]}],
                "Catch": [{"ErrorEquals": ["E"], "Next": "y"}],
                "Iterator": {"StartAt": "z", "States": {"z": {"Type": "Fail"}}},
                "End": True,
            },
            "y": {"Type": "Succeed"},
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == []


def test_ok_min():
    state_machine = {
        "States": {
            "x": {
                "Type": "Map",
                "ItemsPath": "$$.x",
                "Iterator": {"StartAt": "y", "States": {"y": {"Type": "Fail"}}},
                "End": True,
            },
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == []


@pytest.mark.parametrize(
    "reader_config,item_batcher,max_concurrency,tf_count,tf_per",
    [
        (
            {"MaxItems": 1},
            {"MaxItemsPerBatch": 1, "MaxInputBytesPerBatch": 1},
            {"MaxConcurrency": 9999999999},
            {"ToleratedFailureCount": 42},
            {"ToleratedFailurePercentage": 42},
        ),
        (
            {"MaxItemsPath": "$.x"},
            {"MaxItemsPerBatchPath": "$.x", "MaxInputBytesPerBatchPath": "$.x"},
            {"MaxConcurrencyPath": "$.x"},
            {"ToleratedFailureCountPath": "$.x"},
            {"ToleratedFailurePercentagePath": "$.x"},
        ),
    ],
)
def test_ok_2022(reader_config, item_batcher, max_concurrency, tf_count, tf_per):
    state_machine = {
        "States": {
            "x": {
                "Type": "Map",
                "ResultPath": None,
                "ItemSelector": {},
                "ResultSelector": {},
                "ItemsPath": "$$.x",
                **max_concurrency,
                **tf_count,
                **tf_per,
                "Retry": [{"ErrorEquals": ["E"]}],
                "Catch": [{"ErrorEquals": ["E"], "Next": "y"}],
                "ItemProcessor": {
                    "ProcessorConfig": {
                        "Mode": "DISTRIBUTED",
                        "ExecutionType": "EXPRESS",
                    },
                    "StartAt": "z",
                    "States": {"z": {"Type": "Fail"}},
                },
                "ItemReader": {
                    "Resource": "arn:x",
                    "Parameters": {},
                    "ReaderConfig": reader_config,
                },
                "ResultWriter": {
                    "Resource": "arn:x",
                    "Parameters": {},
                },
                "ItemBatcher": {
                    "BatchInput": {},
                    **item_batcher,
                },
                "End": True,
            },
            "y": {"Type": "Succeed"},
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == []


def test_ok_2022_min():
    state_machine = {
        "States": {
            "x": {
                "Type": "Map",
                "ItemsPath": "$$.x",
                "ItemProcessor": {
                    "StartAt": "z",
                    "States": {"z": {"Type": "Fail"}},
                },
                "ItemReader": {"Resource": "arn:x"},
                "ResultWriter": {"Resource": "arn:x"},
                "ItemBatcher": {},
                "End": True,
            },
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == []


def test_iterator_is_not_dict():
    state_machine = {
        "States": {
            "x": {
                "Type": "Map",
                "ResultPath": None,
                "Parameters": {},
                "ResultSelector": {},
                "Retry": [{"ErrorEquals": ["E"]}],
                "Catch": [{"ErrorEquals": ["E"], "Next": "y"}],
                "Iterator": [],
                "End": True,
            },
            "y": {"Type": "Succeed"},
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        "State Machine.States.x.Iterator is [] but should be an Object"
    ]


def test_empty():
    state_machine = {
        "States": {"x": {"Type": "Map"}},
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        'State Machine.States.x does not have required field "Next"',
        "State Machine.States.x does not have required field from "
        '["Iterator", "ItemProcessor"]',
        "No terminal state found in machine at State Machine.States",
    ]


def test_empty_iterator():
    state_machine = {
        "States": {
            "x": {
                "Type": "Map",
                "Iterator": {},
                "End": True,
            },
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        'State Machine.States.x.Iterator does not have required field "States"',
        'State Machine.States.x.Iterator does not have required field "StartAt"',
    ]


@pytest.mark.parametrize("key", ("Iterator", "ItemProcessor"))
def test_redefine_state(key):
    state_machine = {
        "States": {
            "x": {
                "Type": "Map",
                "ItemsPath": "$.x",
                key: {"StartAt": "x", "States": {"x": {"Type": "Fail"}}},
                "End": True,
            },
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        f'State "x", defined at State Machine.States.x.{key}.States, '
        "is also defined at State Machine.States",
    ]


def test_empty_item_reader():
    state_machine = {
        "States": {
            "x": {
                "Type": "Map",
                "ItemsPath": "$$.x",
                "ItemProcessor": {
                    "StartAt": "z",
                    "States": {"z": {"Type": "Fail"}},
                },
                "ItemReader": {},
                "End": True,
            },
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        'State Machine.States.x.ItemReader does not have required field "Resource"'
    ]


def test_non_object_item_reader():
    state_machine = {
        "States": {
            "x": {
                "Type": "Map",
                "ItemsPath": "$$.x",
                "ItemProcessor": {
                    "StartAt": "z",
                    "States": {"z": {"Type": "Fail"}},
                },
                "ItemReader": "x",
                "End": True,
            },
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        'State Machine.States.x.ItemReader is "x" but should be an Object'
    ]


def test_empty_reader_config():
    state_machine = {
        "States": {
            "x": {
                "Type": "Map",
                "ItemsPath": "$$.x",
                "ItemProcessor": {
                    "StartAt": "z",
                    "States": {"z": {"Type": "Fail"}},
                },
                "ItemReader": {"Resource": "arn:x", "ReaderConfig": {}},
                "End": True,
            },
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == []


def test_max_items_zero():
    state_machine = {
        "States": {
            "x": {
                "Type": "Map",
                "ItemsPath": "$$.x",
                "ItemProcessor": {
                    "StartAt": "z",
                    "States": {"z": {"Type": "Fail"}},
                },
                "ItemReader": {"Resource": "arn:x", "ReaderConfig": {"MaxItems": 0}},
                "End": True,
            },
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        "State Machine.States.x.ItemReader.ReaderConfig.MaxItems is 0 but allowed "
        "floor is 0",
    ]


def test_invalid_max_items_path():
    state_machine = {
        "States": {
            "x": {
                "Type": "Map",
                "ItemsPath": "$$.x",
                "ItemProcessor": {
                    "StartAt": "z",
                    "States": {"z": {"Type": "Fail"}},
                },
                "ItemReader": {
                    "Resource": "arn:x",
                    "ReaderConfig": {"MaxItemsPath": "x"},
                },
                "End": True,
            },
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        'State Machine.States.x.ItemReader.ReaderConfig.MaxItemsPath is "x" but '
        "should be a Reference Path",
    ]


def test_both_max_items():
    state_machine = {
        "States": {
            "x": {
                "Type": "Map",
                "ItemsPath": "$$.x",
                "ItemProcessor": {
                    "StartAt": "z",
                    "States": {"z": {"Type": "Fail"}},
                },
                "ItemReader": {
                    "Resource": "arn:x",
                    "ReaderConfig": {"MaxItems": 1, "MaxItemsPath": "$.x"},
                },
                "End": True,
            },
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        "State Machine.States.x.ItemReader.ReaderConfig may have only one of "
        '["MaxItems", "MaxItemsPath"]',
    ]


def test_empty_result_writer():
    state_machine = {
        "States": {
            "x": {
                "Type": "Map",
                "ItemsPath": "$$.x",
                "ItemProcessor": {
                    "StartAt": "z",
                    "States": {"z": {"Type": "Fail"}},
                },
                "ResultWriter": {},
                "End": True,
            },
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        'State Machine.States.x.ResultWriter does not have required field "Resource"'
    ]


def test_non_object_result_writer():
    state_machine = {
        "States": {
            "x": {
                "Type": "Map",
                "ItemsPath": "$$.x",
                "ItemProcessor": {
                    "StartAt": "z",
                    "States": {"z": {"Type": "Fail"}},
                },
                "ResultWriter": "x",
                "End": True,
            },
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        'State Machine.States.x.ResultWriter is "x" but should be an Object'
    ]


def test_result_writer_with_reader_config():
    state_machine = {
        "States": {
            "x": {
                "Type": "Map",
                "ItemsPath": "$$.x",
                "ItemProcessor": {
                    "StartAt": "z",
                    "States": {"z": {"Type": "Fail"}},
                },
                "ResultWriter": {"Resource": "arn:x", "ReaderConfig": {"MaxItems": 1}},
                "End": True,
            },
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        'Field "ReaderConfig" not allowed in State Machine.States.x.ResultWriter'
    ]


def test_non_object_processor_config():
    state_machine = {
        "States": {
            "x": {
                "Type": "Map",
                "ItemsPath": "$$.x",
                "ItemProcessor": {
                    "ProcessorConfig": "x",
                    "StartAt": "z",
                    "States": {"z": {"Type": "Fail"}},
                },
                "End": True,
            },
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        'State Machine.States.x.ItemProcessor.ProcessorConfig is "x" '
        "but should be an Object"
    ]


def test_both_parameters_and_item_selector():
    state_machine = {
        "States": {
            "x": {
                "Type": "Map",
                "Parameters": {},
                "ItemSelector": {},
                "ItemsPath": "$$.x",
                "ItemProcessor": {
                    "StartAt": "z",
                    "States": {"z": {"Type": "Fail"}},
                },
                "End": True,
            },
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        'State Machine.States.x may have only one of ["Parameters", "ItemSelector"]',
    ]


def test_max_items_batcher_zero():
    state_machine = {
        "States": {
            "x": {
                "Type": "Map",
                "ItemsPath": "$$.x",
                "ItemProcessor": {
                    "StartAt": "z",
                    "States": {"z": {"Type": "Fail"}},
                },
                "ItemBatcher": {"MaxItemsPerBatch": 0, "MaxInputBytesPerBatch": 0},
                "End": True,
            },
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        "State Machine.States.x.ItemBatcher.MaxItemsPerBatch is 0 but allowed floor "
        "is 0",
        "State Machine.States.x.ItemBatcher.MaxInputBytesPerBatch is 0 but allowed "
        "floor is 0",
    ]


def test_max_items_batcher_invalid_uri():
    state_machine = {
        "States": {
            "x": {
                "Type": "Map",
                "ItemsPath": "$$.x",
                "ItemProcessor": {
                    "StartAt": "z",
                    "States": {"z": {"Type": "Fail"}},
                },
                "ItemBatcher": {
                    "MaxItemsPerBatchPath": "x",
                    "MaxInputBytesPerBatchPath": "x",
                },
                "End": True,
            },
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        'State Machine.States.x.ItemBatcher.MaxItemsPerBatchPath is "x" but should be '
        "a Reference Path",
        'State Machine.States.x.ItemBatcher.MaxInputBytesPerBatchPath is "x" but '
        "should be a Reference Path",
    ]


@pytest.mark.parametrize(
    "max_concurrency,errors",
    [
        ({"MaxConcurrency": 0}, []),
        (
            {"MaxConcurrency": -1},
            ["State Machine.States.x.MaxConcurrency is -1 but allowed floor is 0"],
        ),
        (
            {"MaxConcurrencyPath": "x"},
            [
                'State Machine.States.x.MaxConcurrencyPath is "x" but should be a '
                "Reference Path"
            ],
        ),
    ],
)
def test_max_concurrency(max_concurrency, errors):
    state_machine = {
        "States": {
            "x": {
                "Type": "Map",
                "ItemsPath": "$$.x",
                "Iterator": {"StartAt": "y", "States": {"y": {"Type": "Fail"}}},
                "End": True,
                **max_concurrency,
            },
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == errors


@pytest.mark.parametrize(
    "tolerated_failure,errors",
    [
        (
            {"ToleratedFailureCount": -1},
            [
                "State Machine.States.x.ToleratedFailureCount is -1 but allowed floor "
                "is 0"
            ],
        ),
        (
            {"ToleratedFailureCount": "x"},
            [
                'State Machine.States.x.ToleratedFailureCount is "x" but should be an '
                "Integer"
            ],
        ),
        (
            {"ToleratedFailurePercentage": -1},
            [
                "State Machine.States.x.ToleratedFailurePercentage is -1 but allowed "
                "floor is 0",
            ],
        ),
        ({"ToleratedFailurePercentage": 100}, []),
        (
            {"ToleratedFailurePercentage": 101},
            [
                "State Machine.States.x.ToleratedFailurePercentage is 101 but allowed "
                "ceiling is 100",
            ],
        ),
        (
            {"ToleratedFailurePercentage": "x"},
            [
                'State Machine.States.x.ToleratedFailurePercentage is "x" but should '
                "be an Integer",
            ],
        ),
        (
            {"ToleratedFailureCount": 0, "ToleratedFailureCountPath": "$.x"},
            [
                'State Machine.States.x may have only one of ["ToleratedFailureCount", '
                '"ToleratedFailureCountPath"]',
            ],
        ),
        (
            {"ToleratedFailurePercentage": 0, "ToleratedFailurePercentagePath": "$.x"},
            [
                "State Machine.States.x may have only one of "
                '["ToleratedFailurePercentage", "ToleratedFailurePercentagePath"]',
            ],
        ),
    ],
)
def test_tolerated_failure(tolerated_failure, errors):
    state_machine = {
        "States": {
            "x": {
                "Type": "Map",
                "ItemsPath": "$$.x",
                "Iterator": {"StartAt": "y", "States": {"y": {"Type": "Fail"}}},
                "End": True,
                **tolerated_failure,
            },
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == errors



import json
from enum import Enum
from typing import Any


class Comparison(str, Enum):
    def __str__(self) -> str:
        return self.value

    STRING_EQUALS = "StringEquals"
    STRING_LESS_THAN = "StringLessThan"
    STRING_GREATER_THAN = "StringGreaterThan"
    STRING_LESS_THAN_EQUALS = "StringLessThanEquals"
    STRING_GREATER_THAN_EQUALS = "StringGreaterThanEquals"
    NUMERIC_EQUALS = "NumericEquals"
    NUMERIC_LESS_THAN = "NumericLessThan"
    NUMERIC_GREATER_THAN = "NumericGreaterThan"
    NUMERIC_LESS_THAN_EQUALS = "NumericLessThanEquals"
    NUMERIC_GREATER_THAN_EQUALS = "NumericGreaterThanEquals"
    BOOLEAN_EQUALS = "BooleanEquals"
    TIMESTAMP_EQUALS = "TimestampEquals"
    TIMESTAMP_LESS_THAN = "TimestampLessThan"
    TIMESTAMP_GREATER_THAN = "TimestampGreaterThan"
    TIMESTAMP_LESS_THAN_EQUALS = "TimestampLessThanEquals"
    TIMESTAMP_GREATER_THAN_EQUALS = "TimestampGreaterThanEquals"
    STRING_EQUALS_PATH = "StringEqualsPath"
    STRING_LESS_THAN_PATH = "StringLessThanPath"
    STRING_GREATER_THAN_PATH = "StringGreaterThanPath"
    STRING_LESS_THAN_EQUALS_PATH = "StringLessThanEqualsPath"
    STRING_GREATER_THAN_EQUALS_PATH = "StringGreaterThanEqualsPath"
    NUMERIC_EQUALS_PATH = "NumericEqualsPath"
    NUMERIC_LESS_THAN_PATH = "NumericLessThanPath"
    NUMERIC_GREATER_THAN_PATH = "NumericGreaterThanPath"
    NUMERIC_LESS_THAN_EQUALS_PATH = "NumericLessThanEqualsPath"
    NUMERIC_GREATER_THAN_EQUALS_PATH = "NumericGreaterThanEqualsPath"
    BOOLEAN_EQUALS_PATH = "BooleanEqualsPath"
    TIMESTAMP_EQUALS_PATH = "TimestampEqualsPath"
    TIMESTAMP_LESS_THAN_PATH = "TimestampLessThanPath"
    TIMESTAMP_GREATER_THAN_PATH = "TimestampGreaterThanPath"
    TIMESTAMP_LESS_THAN_EQUALS_PATH = "TimestampLessThanEqualsPath"
    TIMESTAMP_GREATER_THAN_EQUALS_PATH = "TimestampGreaterThanEqualsPath"
    IS_NULL = "IsNull"
    IS_PRESENT = "IsPresent"
    IS_NUMERIC = "IsNumeric"
    IS_STRING = "IsString"
    IS_BOOLEAN = "IsBoolean"
    IS_TIMESTAMP = "IsTimestamp"
    STRING_MATCHES = "StringMatches"


class StateType(str, Enum):
    def __str__(self) -> str:
        return self.value

    PASS = "Pass"
    SUCCEED = "Succeed"
    FAIL = "Fail"
    TASK = "Task"
    CHOICE = "Choice"
    WAIT = "Wait"
    PARALLEL = "Parallel"
    MAP = "Map"


def to_json(value: Any) -> str:
    return json.dumps(value, ensure_ascii=False)



from typing import List

from ...fields import RESULT_PATH, Field
from ..node import Node


class ResultPathMixin(Node):
    @property
    def optional_fields(self) -> List[Field]:
        return [*super().optional_fields, RESULT_PATH]



from typing import List

from ..fields import DEFAULT, END, NEXT, Field
from .mixins import ChoicesMixin
from .state import State


class ChoiceState(ChoicesMixin, State):
    @property
    def optional_fields(self) -> List[Field]:
        return [*super().optional_fields, DEFAULT]

    @property
    def forbidden_fields(self) -> List[Field]:
        return [NEXT, END, *super().forbidden_fields]



from statelint.linter import Linter


def test_ok():
    state_machine = {
        "States": {
            "x": {
                "Type": "Choice",
                "Choices": [{"Variable": "$", "IsNull": True, "Next": "y"}],
                "Default": "z",
            },
            "y": {"Type": "Succeed"},
            "z": {"Type": "Succeed"},
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == []


def test_ok_without_default():
    state_machine = {
        "States": {
            "x": {
                "Type": "Choice",
                "Choices": [{"Variable": "$", "IsNull": True, "Next": "y"}],
            },
            "y": {"Type": "Succeed"},
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == []


def test_empty_choices():
    state_machine = {
        "States": {
            "x": {
                "Type": "Choice",
                "Choices": [],
                "Default": "z",
            },
            "z": {"Type": "Succeed"},
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        "State Machine.States.x.Choices is empty, non-empty required"
    ]


def test_next_end():
    state_machine = {
        "States": {
            "x": {
                "Type": "Choice",
                "Choices": [{"Variable": "$", "IsNull": True, "Next": "y"}],
                "Next": "z",
                "End": True,
            },
            "y": {"Type": "Succeed"},
            "z": {"Type": "Succeed"},
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        'State Machine.States.x has forbidden field "Next"',
        'State Machine.States.x has forbidden field "End"',
    ]


def test_not_numeric():
    state_machine = {
        "States": {
            "x": {
                "Type": "Choice",
                "Choices": [{"Variable": "$", "NumericEquals": "1", "Next": "y"}, None],
                "Default": "z",
            },
            "y": {"Type": "Succeed"},
            "z": {"Type": "Succeed"},
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        "State Machine.States.x.Choices[1] is null but should be an Object",
        'State Machine.States.x.Choices[0].NumericEquals is "1" but should be numeric',
    ]


def test_non_list_choices():
    state_machine = {
        "States": {
            "x": {"Type": "Choice", "Choices": {}, "Default": "y"},
            "y": {"Type": "Succeed"},
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        "State Machine.States.x.Choices is {} but should be an Array"
    ]



from statelint.linter import Linter


def test_ok():
    state_machine = {
        "States": {
            "x": {
                "Type": "Parallel",
                "ResultPath": None,
                "Parameters": {},
                "ResultSelector": {},
                "Retry": [{"ErrorEquals": ["E"]}],
                "Catch": [{"ErrorEquals": ["E"], "Next": "y"}],
                "Branches": [{"StartAt": "z", "States": {"z": {"Type": "Fail"}}}],
                "End": True,
            },
            "y": {"Type": "Succeed"},
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == []


def test_empty():
    state_machine = {
        "States": {"x": {"Type": "Parallel"}},
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        'State Machine.States.x does not have required field "Next"',
        'State Machine.States.x does not have required field "Branches"',
        "No terminal state found in machine at State Machine.States",
    ]


def test_empty_branches():
    state_machine = {
        "States": {
            "x": {
                "Type": "Parallel",
                "Branches": [],
                "End": True,
            },
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == []


def test_no_transition():
    state_machine = {
        "States": {
            "x": {
                "Type": "Parallel",
                "ResultPath": None,
                "Parameters": {},
                "ResultSelector": {},
                "Retry": [{"ErrorEquals": ["E"]}],
                "Catch": [{"ErrorEquals": ["E"], "Next": "y"}],
                "Branches": [
                    {
                        "StartAt": "z",
                        "States": {"z": {"Type": "Fail"}, "z2": {"Type": "Fail"}},
                    }
                ],
                "End": True,
            },
            "y": {"Type": "Succeed"},
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        "No transition found to state State Machine.States.x.Branches[0].z2"
    ]
